// Frickin PPTP Proxy, Copyright (c) 2004-06 Martin Akesson <@placid.tv>
// 
// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to
// the following conditions:
// 
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

#include <unistd.h>
#include <sys/fcntl.h>
#include <sys/poll.h>
#include <signal.h>
#include <pthread.h>
#include <list>
#include <iostream>

#include "configuration.hpp"
#include "listener.hpp"
#include "session.hpp"
#include "grehandler.hpp"
#include "exception.hpp"
#include "logger.hpp"
#include "nat.hpp"
#include "util.hpp"

using namespace std;

struct Poll
{
    struct pollfd* p_PollList;
    Entity** p_EntityList;
    unsigned int p_Count;
    unsigned int p_Start;
    unsigned int p_Size;
};

char* g_ConfigFile = "frickin.conf";
const char* g_AppName = "Frickin";
const int g_AppVersionHi = 2;
const int g_AppVersionLo = 0;

int g_PendingConnects = 0;
int g_PendingSessions = 0;
pthread_mutex_t g_ConnectMutex;
pthread_mutex_t g_SessionMutex;
list<Session*> g_NewSessions;

void nullSigHandler(int a_Sig)
{
    //cout << "SIGALARM!" << std::endl;
}
/* connectHandler() function is used by helper threads
 * to offload connecting clients from the main thread. */
void* connectHandler(void* a_Arg)
{
    Session* session = static_cast<Session*>(a_Arg);

    if (session)
    {
	try
	{
	    session->connectServer();
	    pthread_mutex_lock(&g_SessionMutex);
	    g_NewSessions.push_back(session);
	    g_PendingSessions++;
	    pthread_mutex_unlock(&g_SessionMutex);
	    usleep(500000);
	    //	    kill(getpid(), SIGALRM);	// FIXME ???
	}
	catch (NetworkException &e)
	{
	    // TODO, it would be nice to know what server caused the error
	    Logger::getLogger()->log(LOG_ERR, "Failed to connect to server (%s)", e.errorMessage());
	    delete session;
	}
    }

    pthread_mutex_lock(&g_ConnectMutex);
    g_PendingConnects--;
    pthread_mutex_unlock(&g_ConnectMutex);
    return NULL;
}

// I really dont like this solution..  still thinking of something better
// I'd like to only handle sessions and no entities in main()
void pollInit(struct Poll* a_PollInfo, list<Session*> a_SessionList)
{
    int pollIndex = a_PollInfo->p_Start;
    Entity** entityList = NULL;
    int listSize = 0;
    int socketCount;

    a_PollInfo->p_Count = a_PollInfo->p_Start;
    for (list<Session*>::iterator sessionIter = a_SessionList.begin(); sessionIter != a_SessionList.end(); sessionIter++)
    {
	int i;
	socketCount = (*sessionIter)->getEntityList(entityList, listSize);
	
	// WARNING: pollIndex may overrun..  fix this crap
	for (i = 0; i < socketCount; i++, pollIndex++)
	{
	    a_PollInfo->p_EntityList[pollIndex] = entityList[i];
	    a_PollInfo->p_PollList[pollIndex].fd = entityList[i]->m_Socket;
	    a_PollInfo->p_PollList[pollIndex].events = POLLIN;
	    a_PollInfo->p_Count++;
	}
    }
    if (entityList) free(entityList);
}

int main(int argc, char** argv)
{
    Configuration* cfg = NULL;
    unsigned int logFlags = LOGGER_SYSLOG;
    bool foregroundMode = 0;
    int ret;
    list<Listener*> listeners;
    GreHandler* greHandler = NULL;
    Logger* logger;
    struct Poll pollInfo;

    // Parse command line options and apply settings
    while (-1 < (ret = getopt(argc, argv, "c:dfos")))
    {
	switch (ret)
	{
	    case 'c': // Configuration file
		g_ConfigFile = optarg;
		break;
	    case 'd': // Enable debug logging
		logFlags |= LOGGER_DEBUG;
		break;
	    case 'f': // Run in foreground (do not daemonize)
		foregroundMode = 1;
		logFlags |= LOGGER_STDOUT;
		break;
	    case 'o': // Enable standard out logging
		logFlags |= LOGGER_STDOUT;
		break;
	    case 's': // Disable syslog logging
		logFlags &= (-1 ^ LOGGER_SYSLOG);
		break;
	}
    }

    srandomdev();
    signal(SIGPIPE, SIG_IGN);
    signal(SIGALRM, nullSigHandler);

    // Instantiate the logger
    logger = new Logger(g_AppName, LOG_DAEMON, logFlags);

    // Unless foreground mode is specified we fork off a child to daemonize
    if (!foregroundMode)
    {
	int pid;

	pid = fork();
	if (pid)
	{
	    if (-1 == pid)
	    {
		logger->log(LOG_ERR, "%s v%d.%dBETA1, unable to daemonize (%s)",
			g_AppName, g_AppVersionHi, g_AppVersionLo, strerror(errno));
	    }
	    else
	    {
		logger->log(LOG_INFO, "%s v%d.%dBETA1, running as daemon with pid %d",
			g_AppName, g_AppVersionHi, g_AppVersionLo, pid);
	    }
	    exit(0);
	}
    }
    else
    {
	logger->log(LOG_INFO, "%s v%d.%dBETA1, running in foreground",
		g_AppName, g_AppVersionHi, g_AppVersionLo);
    }

    try
    {
	// Initialize the GRE packet handler, this must be done with root privileges
	greHandler = new GreHandler();
	pollInfo.p_Count = 1;

	// Initialize the nat handler (Open /dev/pf when using OpenBSD packet filter)
	natInit();

	// Load the configuration file
	cfg = new Configuration(g_ConfigFile);

	// All operations that need root privileges must be done prior to
	// this line.  We will now drop all root privileges and assume
	// the identity specified in the configuration file.
	dropPrivileges(cfg->user(), cfg->group());

	pollInfo.p_Count += cfg->numListeners();
	pollInfo.p_Start = pollInfo.p_Count;
	pollInfo.p_Size = pollInfo.p_Count;
	pollInfo.p_Size += 100; // TODO, client.max * 2
	pollInfo.p_PollList = reinterpret_cast<struct pollfd*>(calloc(pollInfo.p_Size, sizeof(struct pollfd)));
	pollInfo.p_EntityList = reinterpret_cast<Entity**>(calloc(pollInfo.p_Size, sizeof(Entity*)));
	if (NULL == pollInfo.p_PollList) throw MemoryException(errno);

	// Set up the gre poll struct
	pollInfo.p_PollList[0].fd = greHandler->greSocket();
	pollInfo.p_PollList[0].events = POLLIN;

	// Initialize all listeners and their poll structs
	if (cfg->numListeners() == 0)
	{
	    logger->log(LOG_ERR, "No listeners configured, terminating!");
	    return -1;
	}
	else
	{
	    int i;

	    for (i = 0; i < cfg->numListeners(); i++)
	    {
		Listener* newListener;

		newListener = cfg->newListener(i);
		listeners.push_back(newListener);
		pollInfo.p_PollList[i + 1].fd = newListener->m_Socket;
		pollInfo.p_PollList[i + 1].events = POLLIN;
		logger->log(LOG_INFO, "Listener on %s:%d, %s", newListener->address(), newListener->port(), newListener->name());
	    }
	}
    }
    catch (FileIOException)
    {
	logger->log(LOG_ERR, "Could not open configuration file '%s'", g_ConfigFile);
	return -1;
    }
    catch (ParseException& e)
    {
	logger->log(LOG_ERR, "Syntax error in configuration file (Line: %d, %s", e.getLine(), e.getError());
	return -1;
    }
    catch (SettingNotFoundException& e)
    {
	logger->log(LOG_ERR, "Aborting, missing configuration");
	return -1;
    }
    catch (SettingTypeException& e)
    {
	logger->log(LOG_ERR, "Type error in configuration");
	return -1;
    }
    catch (NetworkException& e)
    {
	logger->log(LOG_ERR, "Network error (%s)", e.errorMessage());
	return -1;
    }

    pthread_mutex_init(&g_ConnectMutex, NULL);
    pthread_mutex_init(&g_SessionMutex, NULL);
    list<Session*> sessionList;
    for (;;)
    {
	int readySockets;

	readySockets = poll(pollInfo.p_PollList, pollInfo.p_Count, 500);

	if (0 < readySockets)
	{
	    if (pollInfo.p_PollList[0].revents)
	    { // GRE socket
		list<Call*> callList = Call::callList();
		struct sockaddr_in sockAddr;
		u_int16_t callID = 0;
		int greLength = 0;

		if (greHandler->recvFrom(&sockAddr, callID, greLength))
		{
		    // TODO, Handle the error
		    // Most likely we should just terminate..
		}

		if (0 < greLength)
		{
		    for (list<Call*>::iterator callIter = callList.begin(); callIter != callList.end(); callIter++)
		    {
			if (**callIter == callID)
			{
			    Call* theCall = *callIter;
			    if (theCall->m_ServerAddr.sin_addr.s_addr == sockAddr.sin_addr.s_addr)
			    { // From server
				if (greHandler->sendTo(&(theCall->m_SpoofAddr), &(theCall->m_ClientAddr), theCall->clientID(), greLength))
				{
				    // TODO, Handle the error
				    // Most likely we should just terminate..
				}
			    }
			    else
			    { // From client
				if (greHandler->sendTo(&(theCall->m_ServerAddr), theCall->serverID(), greLength))
				{
				    // TODO, Handle the error
				    // Most likely we should just terminate..
				}
			    }
			}
		    }
		}
	    }

	    // Unless the current number of pending connections is below
	    // specified threshold we should not process any new connections.
	    if (g_PendingConnects < 10)
	    {
		int pollIndex = 1;
		for (list<Listener*>::iterator listenerIter = listeners.begin(); listenerIter != listeners.end(); listenerIter++, pollIndex++)
		{
		    if (pollInfo.p_PollList[pollIndex].revents)
		    {
			    Session* newSession = NULL;
			    pthread_t newThread;
			    pthread_attr_t threadAttr;

			    if (!Session::getSession((*listenerIter)->accept(), newSession))
			    {
				// Helper threads will be created for each connecting
				// client so they dont lock up the main thread.
				pthread_mutex_lock(&g_ConnectMutex);
				pthread_attr_init(&threadAttr);
				pthread_attr_setdetachstate(&threadAttr, true);
				pthread_create(&newThread, &threadAttr, connectHandler, static_cast<void*>(newSession));
				g_PendingConnects++;
				pthread_mutex_unlock(&g_ConnectMutex);
			    }
			    else
			    {
				if (newSession) delete newSession;
				logger->log(LOG_ERR, "Failed to connect to server..");  // TODO, needs more info
			    }
		    }
		}
	    }

	    // Process each entity sequentially by iterating the entity list
	    for (unsigned int pollIndex = pollInfo.p_Start; pollIndex < pollInfo.p_Count; pollIndex++)
	    {
		Entity* theEntity;
		theEntity = (pollInfo.p_EntityList[pollIndex]);
		if (theEntity->m_Session)
		{
		    if (pollInfo.p_PollList[pollIndex].revents)
		    {
			if (theEntity->process())
			{
			    Session* theSession;
			    // Detatch the client/server from this session
			    theSession = theEntity->detatch();
			    delete theEntity;
			    if (theSession)
			    {
				// Session was deleted, remove from the session list
				sessionList.erase(find(sessionList.begin(), sessionList.end(), theSession));
			    }
			}
		    }
		}
		else
		{
		    // This should never happen..
		    delete theEntity;
		}
	    }
	}

	// If we have new sessions ready we should add them to the session list
	if (g_PendingSessions)
	{
	    pthread_mutex_lock(&g_SessionMutex);
	    for (list<Session*>::iterator sessionIter = g_NewSessions.begin(); sessionIter != g_NewSessions.end();)
	    {
		sessionList.push_back(*sessionIter);
		sessionIter = g_NewSessions.erase(sessionIter);
		g_PendingSessions--;
	    }
	    pthread_mutex_unlock(&g_SessionMutex);
	}
	pollInit(&pollInfo, sessionList);
    }

    return ret;
}
