// Frickin PPTP Proxy, Copyright (c) 2004-06 Martin Akesson <@placid.tv>
// 
// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to
// the following conditions:
// 
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include "rfc2637.hpp"

const char *pptpMsgNames[] =
{
	NULL,
	"Start-Control-Connection-Request",
	"Start-Control-Connection-Reply",
	"Stop-Control-Connection-Request",
	"Stop-Control-Connection-Reply",
	"Echo-Request",
	"Echo-Reply",
	"Outgoing-Call-Request",
	"Outgoing-Call-Reply",
	"Incoming-Call-Request",
	"Incoming-Call-Reply",
	"Incoming-Call-Connected",
	"Call-Clear-Request",
	"Call-Disconnect-Notify",
	"WAN-Error-Notify",
	"Set-Link-Info"
};

const char *pptpErrorNames[] =
{
	"No error",
	"Not connected",
	"Bad format",
	"Bad value",
	"No resource",
	"Bad CALL-ID",
	"PAC error"
};

int build_SCCRQ(void* a_Buffer, int a_ProtocolHi, int a_ProtocolLo, int a_FramingCap,
		int a_BearerCap, int a_MaxChannels, int a_FirmwareRev,
		char* a_Hostname, char* a_Vendor)
{
	struct sccrq_msg* sccrq = reinterpret_cast<struct sccrq_msg*>(a_Buffer);

	memset(a_Buffer, 0, sizeof(struct sccrq_msg));
	sccrq->length = htons(sizeof(struct sccrq_msg));
	sccrq->type = htons(PPTP_CTRL);
	sccrq->magicCookie = htonl(MAGIC_COOKIE);
	sccrq->ctrlType = htons(SCCRQ);
	sccrq->protocolHi = a_ProtocolHi;
	sccrq->protocolLo = a_ProtocolLo;
	sccrq->framingCap = htonl(a_FramingCap);
	sccrq->bearerCap = htonl(a_BearerCap);
	sccrq->maxChannels = htons(a_MaxChannels);
	sccrq->firmwareRev = htons(a_FirmwareRev);
	if (a_Hostname) strncpy(sccrq->hostname, a_Hostname, 64);
	if (a_Vendor) strncpy(sccrq->vendor, a_Vendor, 64);
	return sizeof(struct sccrq_msg);
}

int build_SCCRP(void* a_Buffer, int a_ProtocolHi, int a_ProtocolLo, int a_Result, int a_Error,
		unsigned int a_FramingCap, unsigned int a_BearerCap, int a_MaxChannels,
		int a_FirmwareRev, char* a_Hostname, char* a_Vendor)
{
	struct sccrp_msg *sccrp = reinterpret_cast<struct sccrp_msg*>(a_Buffer);

	memset(a_Buffer, 0, sizeof(struct sccrp_msg));
	sccrp->length = htons(sizeof(struct sccrp_msg));
	sccrp->type = htons(PPTP_CTRL);
	sccrp->magicCookie = htonl(MAGIC_COOKIE);
	sccrp->ctrlType = htons(SCCRP);
	sccrp->protocolHi = a_ProtocolHi;
	sccrp->protocolLo = a_ProtocolLo;
	sccrp->result = a_Result;
	sccrp->error = a_Error;
	sccrp->framingCap = htonl(a_FramingCap);
	sccrp->bearerCap = htonl(a_BearerCap);
	sccrp->maxChannels = htons(a_MaxChannels);
	sccrp->firmwareRev = htons(a_FirmwareRev);
	if (a_Hostname) strncpy(sccrp->hostname, a_Hostname, 64);
	if (a_Vendor) strncpy(sccrp->vendor, a_Vendor, 64);
	return sizeof(struct sccrp_msg);
}

int build_ECORQ(void* a_Buffer, int a_Identifier)
{
	struct ecorq_msg *ecorq = reinterpret_cast<struct ecorq_msg*>(a_Buffer);

	memset(a_Buffer, 0, sizeof(struct ecorq_msg));
	ecorq->length = htons(sizeof(struct ecorq_msg));
	ecorq->type = htons(PPTP_CTRL);
	ecorq->magicCookie = htonl(MAGIC_COOKIE);
	ecorq->ctrlType = htons(ECORQ);
	ecorq->identifier = htonl(a_Identifier);
	return sizeof(struct ecorq_msg);
}

int build_ECORP(void* a_Buffer, int a_Identifier, int a_Result, int a_Error)
{
	struct ecorp_msg *ecorp = reinterpret_cast<struct ecorp_msg*>(a_Buffer);

	memset(a_Buffer, 0, sizeof(struct ecorp_msg));
	ecorp->length = htons(sizeof(struct ecorp_msg));
	ecorp->type = htons(PPTP_CTRL);
	ecorp->magicCookie = htonl(MAGIC_COOKIE);
	ecorp->ctrlType = htons(ECORP);
	ecorp->identifier = htonl(a_Identifier);
	ecorp->result = a_Result;
	ecorp->error = a_Error;
	return sizeof(struct ecorp_msg);
}

int build_OCRQ(void* a_Buffer, int a_CallId, unsigned int a_MinBps, unsigned int a_MaxBps,
		unsigned int a_BearerType, unsigned int a_FramingType, int a_RecvWinSz,
		int a_ProcDelay, char* a_PhoneNumber, char* a_Subaddress)
{
	struct ocrq_msg *ocrq = reinterpret_cast<struct ocrq_msg*>(a_Buffer);

	memset(a_Buffer, 0, sizeof(struct ocrq_msg));
	ocrq->length = htons(sizeof(struct ocrq_msg));
	ocrq->type = htons(PPTP_CTRL);
	ocrq->magicCookie = htonl(MAGIC_COOKIE);
	ocrq->ctrlType = htons(OCRQ);
	ocrq->callId = a_CallId;
	ocrq->minBps = htonl(a_MinBps);
	ocrq->maxBps = htonl(a_MaxBps);
	ocrq->bearerType = htonl(a_BearerType);
	ocrq->framingType = htonl(a_FramingType);
	ocrq->recvWinSz = htons(a_RecvWinSz);
	ocrq->procDelay = htons(a_ProcDelay);
	if (a_PhoneNumber) strncpy(ocrq->phoneNumber, a_PhoneNumber, 64);
	if (a_Subaddress) strncpy(ocrq->subaddress, a_Subaddress, 64);
	return sizeof(struct ocrq_msg);
}

int build_OCRP(void* a_Buffer, int a_CallId, int a_PeerCallId, int a_Result, int a_Error,
		int a_Cause, unsigned int a_Speed, int a_RecvWinSz, int a_ProcDelay, unsigned int a_ChannelId)
{
	struct ocrp_msg *ocrp = reinterpret_cast<struct ocrp_msg*>(a_Buffer);

	memset(a_Buffer, 0, sizeof(struct ocrp_msg));
	ocrp->length = htons(sizeof(struct ocrp_msg));
	ocrp->type = htons(PPTP_CTRL);
	ocrp->magicCookie = htonl(MAGIC_COOKIE);
	ocrp->ctrlType = htons(OCRP);
	ocrp->callId = a_CallId;
	ocrp->peerCallId = a_PeerCallId;
	ocrp->result = a_Result;
	ocrp->error = a_Error;
	ocrp->cause = htons(a_Cause);
	ocrp->speed = htonl(a_Speed);
	ocrp->recvWinSz = htons(a_RecvWinSz);
	ocrp->procDelay = htons(a_ProcDelay);
	ocrp->channelId = htonl(a_ChannelId);
	return sizeof(struct ocrp_msg);
}

/*
int build_TCCRQ(int a_reason)
{
	struct tccrq_msg *tccrq = reinterpret_cast<struct tccrq_msg*>(a_buffer);

	memset(a_buffer, 0, sizeof(struct tccrq_msg));
	tccrq->length = htons(sizeof(struct tccrq_msg));
	tccrq->type = htons(PPTP_CTRL);
	tccrq->magic_cookie = htonl(MAGIC_COOKIE);
	tccrq->ctrl_type = htons(TCCRQ);
	tccrq->reason = htons(a_reason);
	return send(a_socket, a_buffer, sizeof(struct tccrq_msg), 0);
}
*/

int build_TCCRP(void* a_Buffer, int a_Result, int a_Error)
{
	struct tccrp_msg* tccrp = reinterpret_cast<struct tccrp_msg*>(a_Buffer);

	memset(a_Buffer, 0, sizeof(struct tccrp_msg));
	tccrp->length = htons(sizeof(struct tccrp_msg));
	tccrp->type = htons(PPTP_CTRL);
	tccrp->magicCookie = htonl(MAGIC_COOKIE);
	tccrp->ctrlType = htons(TCCRP);
	tccrp->result = a_Result;
	tccrp->error = a_Error;
	return sizeof(struct tccrp_msg);
}

int build_CCRQ(void* a_Buffer, int a_CallId)
{
	struct ccrq_msg *ccrq = reinterpret_cast<struct ccrq_msg*>(a_Buffer);

	memset(a_Buffer, 0, sizeof(struct ccrq_msg));
	ccrq->length = htons(sizeof(struct ccrq_msg));
	ccrq->type = htons(PPTP_CTRL);
	ccrq->magicCookie = htonl(MAGIC_COOKIE);
	ccrq->ctrlType = htons(CCRQ);
	ccrq->callId = a_CallId;
	return sizeof(struct ccrq_msg);
}

int build_CDN(void* a_Buffer, int a_CallId, int a_Result, int a_Error, int a_Cause, char* a_Stats)
{
	struct cdn_msg *cdn = reinterpret_cast<struct cdn_msg*>(a_Buffer);

	memset(a_Buffer, 0, sizeof(struct cdn_msg));
	cdn->length = htons(sizeof(struct cdn_msg));
	cdn->type = htons(PPTP_CTRL);
	cdn->magicCookie = htonl(MAGIC_COOKIE);
	cdn->ctrlType = htons(CDN);
	cdn->callId = a_CallId;
	cdn->result = a_Result;
	cdn->error = a_Error;
	cdn->cause = htons(a_Cause);
	if (a_Stats) memcpy(cdn->stats, a_Stats, 128);
	return sizeof(struct cdn_msg);
}
