// Frickin PPTP Proxy, Copyright (c) 2004-06 Martin Akesson <@placid.tv>
// 
// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to
// the following conditions:
// 
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

#include <iostream>
#include <unistd.h>
#include <strings.h>
#include <arpa/inet.h>
#include <errno.h>

#include "session.hpp"
#include "logger.hpp"
#include "util.hpp"
#include "rfc2637.hpp"
#include "nat.hpp"

using namespace std;

list<Session*> Session::s_SessionList;

Session::Session(Server* a_Server, Client* a_Client)
{
    m_Server = a_Server;
    a_Server->m_Session = this;
    attatch(a_Client);

    memset(m_Hostname, 0, 64);
    memset(m_Vendor, 0, 64);

    s_SessionList.push_back(this);
    DEBUGF(("Session::Session = %p\n", this));
}

Session::~Session()
{
    for (list<Client*>::iterator clientIter = m_ClientList.begin(); clientIter != m_ClientList.end();)
    {
	delete *clientIter;
	clientIter = m_ClientList.erase(clientIter);
    }
    m_Server->m_Session = NULL;

    s_SessionList.erase(find(s_SessionList.begin(), s_SessionList.end(), this));
    DEBUGF(("Session::~Session = %p\n", this));
}

int Session::getSession(const int a_Socket, Session*& a_Session)
{
    Client* newClient = NULL;
    Server* newServer = NULL;
    struct sockaddr_in sourceAddr4;
    struct sockaddr_in serverAddr4;
    struct sockaddr_in targetAddr4;
    socklen_t sockLen;

    if (0 > a_Socket) throw NetworkException(errno);

    a_Session = NULL;
    sockLen = sizeof(sourceAddr4);
    getpeername(a_Socket, reinterpret_cast<struct sockaddr*>(&sourceAddr4), &sockLen);
    sockLen = sizeof(serverAddr4);
    getsockname(a_Socket, reinterpret_cast<struct sockaddr*>(&serverAddr4), &sockLen);

    if (natLook4(a_Socket, sourceAddr4, targetAddr4))
    {
	// Natlook failed
    }
    else
    {
	serverAddr4 = targetAddr4;
    }

    try
    {
	newClient = new Client(a_Socket, sourceAddr4, serverAddr4, targetAddr4);
	newServer = new Server(newClient->server());

	for (list<Session*>::iterator session_iter = s_SessionList.begin(); session_iter != s_SessionList.end(); session_iter++)
	{
	    if (**session_iter == newServer)
	    {
		// Server connection already exists
		(*session_iter)->attatch(newClient);
		delete newServer;
		return 0;
	    }
	}

	if (NULL == a_Session)
	{
	    // Create new session
	    a_Session = new Session(newServer, newClient);
	}
    }
    catch (...)
    {
	if (newClient) delete newClient;
	if (newServer) delete newServer;
    }

    return (NULL == a_Session);
}

void Session::attatch(Client* a_Client)
{
    m_ClientList.push_back(const_cast<Client*>(a_Client));
    a_Client->m_Session = this;
}

Session* Session::detatch(Client* a_Client)
{
    list<Client*>::iterator clientIter = NULL;

    clientIter = find(m_ClientList.begin(), m_ClientList.end(), a_Client);
    if (clientIter != m_ClientList.end())
    {
	m_ClientList.erase(clientIter);
	a_Client->m_Session = NULL;
    }

    return this;
}

void Session::connectServer() throw(NetworkException)
{
    m_Server->connect();
}

int Session::getEntityList(Entity**& a_EntityList, int& a_ListSize)
{
    int numEntities = 1 + m_ClientList.size();

    if (a_ListSize < numEntities)
    {
	Entity** tmpEntities;

	tmpEntities = static_cast<Entity**>(realloc(a_EntityList, numEntities * sizeof(Entity*)));
	if (tmpEntities)
	{
	    a_ListSize = numEntities;
	    a_EntityList = tmpEntities;
	}
	else
	{
	    return -1;
	}
    }

    a_EntityList[0] = m_Server;

    int i = 1;
    for (list<Client*>::iterator clientIter = m_ClientList.begin(); clientIter != m_ClientList.end(); clientIter++, i++)
    {
	a_EntityList[i] = (*clientIter);
    }

    return numEntities;
}

int Session::process(Server* a_Server)
{
    void* buffer = m_Server->m_Buffer;
    int& bufferUse = m_Server->m_BufferUse;
    struct pptp_msg* pptp = reinterpret_cast<struct pptp_msg*>(buffer);
    int ret = 0;

    switch (m_Server->m_CtrlState)
    {
	case SERVER_WAITSTOPREPLY: // Server has requested a tear down of the connection
	    // This should never be reached
	    break;
	case SERVER_ESTABLISHED: // Control connection has been established
	    switch (ntohs(pptp->ctrlType))
	    {
		case TCCRQ: // Stop-Control-Connection-Request [RFC2637 2.3]
		    if ((bufferUse == ntohs(pptp->length)) && (bufferUse == sizeof(struct tccrq_msg)) && (MAGIC_COOKIE == ntohl(pptp->magicCookie)))
		    {
			// Send the stop request to all connected clients
			for (list<Client*>::iterator client_iter = m_ClientList.begin(); client_iter != m_ClientList.end(); client_iter++)
			{
			    (*client_iter)->write(buffer, bufferUse);
			}
			// Prepare "OK" reply and send it to the server
			bufferUse = build_TCCRP(buffer, TCC_OK, 0);
			m_Server->write(buffer, bufferUse);
			ret = -1;
		    }
		    break;
		case ECORQ: // Echo-Request [RFC2637 2.5]
		    if ((bufferUse == ntohs(pptp->length)) && (bufferUse == sizeof(struct ecorq_msg)) && (MAGIC_COOKIE == ntohl(pptp->magicCookie)))
		    {
			// FIXME, Verify all clients last update time and detatch those who are dead
			struct ecorp_msg* ecorp = reinterpret_cast<struct ecorp_msg*>(buffer);

			bufferUse = build_ECORP(buffer, ecorp->identifier, ECO_OK, 0);
			m_Server->write(buffer, bufferUse);
		    }
		    break;
		case ECORP: // Echo-Reply [RFC2637 2.6]
		    // TODO, Not implemented
		    break;
		case OCRP: // Outgoing-Call-Reply [RFC2637 2.8]
		    if ((bufferUse == ntohs(pptp->length)) && (bufferUse == sizeof(struct ocrp_msg)) && (MAGIC_COOKIE == ntohl(pptp->magicCookie)))
		    {
			// If we know of this call we should replace the ServerID with
			// the ProxyID and relay to the client, otherwise we silently
			// drop the message.
			struct ocrp_msg* ocrp = reinterpret_cast<struct ocrp_msg*>(buffer);

			for (list<Call*>::iterator call_iter = m_CallList.begin(); call_iter != m_CallList.end(); call_iter++)
			{
			    if (ocrp->peerCallId == (*call_iter)->m_ProxyID)
			    {
				(*call_iter)->m_ServerID = ocrp->callId;
				ocrp->callId = (*call_iter)->m_ProxyID;
				ocrp->peerCallId = (*call_iter)->m_ClientID;
				(*call_iter)->m_Client->write(buffer, bufferUse);
				break;
			    }
			}
		    }
		    break;
		case CCRQ: // Call-Clear-Request [RFC2637 2.12]
		    if ((bufferUse == ntohs(pptp->length)) && (bufferUse == sizeof(struct ccrq_msg)) && (MAGIC_COOKIE == ntohl(pptp->magicCookie)))
		    {
			struct ccrq_msg* ccrq = reinterpret_cast<struct ccrq_msg*>(buffer);

			// Search the call list for a matching Call ID
			for (list<Call*>::iterator call_iter = m_CallList.begin(); call_iter != m_CallList.end(); call_iter++)
			{
			    if ((*call_iter)->m_ServerID == ccrq->callId)
			    {
				// Replace the ServerID with our ProxyID and send to the client.
				ccrq->callId = (*call_iter)->m_ProxyID;
				(*call_iter)->m_Client->write(buffer, bufferUse);

				// Respond immediately to the server with a Call Disconnect Notify
				bufferUse = build_CDN(buffer, ccrq->callId, CDN_REQUEST, NONE, 0, NULL);
				m_Server->write(buffer, bufferUse);

				// Delete the call and also remove it from the call list
				delete (*call_iter);
				m_CallList.erase(call_iter);
				break;
			    }
			}
		    }
		    break;
		case CDN: // Call-Disconnect-Notify [RFC2637 2.13]
		    if ((bufferUse == ntohs(pptp->length)) && (bufferUse == sizeof(struct cdn_msg)) && (MAGIC_COOKIE == ntohl(pptp->magicCookie)))
		    {
			// TODO: implement server CDN messages
#if 0
			struct cdn_msg* cdn = reinterpret_cast<struct cdn_msg*>(mBuffer);

			// If we know of this call we should replace the ServerID with
			// the ProxyID and relay to the client, otherwise we silently
			// drop the message.
			for (list<Call*>::iterator call_iter = mCallList.begin(); call_iter != mCallList.end(); call_iter++)
			{
			    if ((*call_iter)->serverID() == cdn->callId)
			    {
				// We dont proxy call disconnect notifications yet.
				// Instead we just replace the server id with our
				// proxy id and relay to the client.
				cdn->callId = (*call_iter)->proxyID();
				(*call_iter)->mClient->write(mBuffer, mBufferUz);
				break;
			    }
			}
#endif
		    }
		    break;
	    }
	    break;
	case SERVER_IDLE: // Server is in idle state waiting for client SCCRQ
	    switch (ntohs(pptp->ctrlType))
	    {
		case SCCRP: // Start-Control-Connection-Reply [RFC2637 2.2]
		    {
			struct sccrp_msg* sccrp = reinterpret_cast<struct sccrp_msg*>(buffer);

			if (ntohs(sccrp->ctrlType) == SCCRP)
			{
			    if (bufferUse == ntohs(sccrp->length) && (bufferUse == sizeof(struct sccrp_msg)) && (MAGIC_COOKIE == ntohl(pptp->magicCookie)))
			    {
				if (sccrp->result == SCC_OK)
				{
				    m_FramingCap = ntohl(sccrp->framingCap);
				    m_BearerCap = ntohl(sccrp->bearerCap);
				    m_MaxChannels = ntohs(sccrp->maxChannels);
				    m_FirmwareRev = ntohs(sccrp->firmwareRev);
				    if (m_Hostname[0] == 0) memcpy(&m_Hostname, sccrp->hostname, 64);
				    if (m_Vendor[0] == 0) memcpy(&m_Vendor, sccrp->vendor, 64);
				    m_Server->m_CtrlState = SERVER_ESTABLISHED;
				}
				bufferUse = build_SCCRP(buffer, 1, 0, SCC_OK, 0, m_FramingCap, m_BearerCap, m_MaxChannels, m_FirmwareRev, m_Hostname, m_Vendor);
				for (list<Client*>::iterator client_iter = m_ClientList.begin(); client_iter != m_ClientList.end(); client_iter++)
				{
				    if ((*client_iter)->m_CtrlState == CLIENT_WAITCTRLREPLY)
				    {
					(*client_iter)->m_CtrlState = CLIENT_ESTABLISHED;
					(*client_iter)->write(buffer, bufferUse);
					break;
				    }
				}
			    }
			}
			break;
		    }
		default:
		    break;
	    }
	    break;
    }

    return ret;
}

int Session::process(Client* a_Client)
{
    void* buffer = a_Client->m_Buffer;
    int& bufferUse = a_Client->m_BufferUse;
    struct pptp_msg* pptp = reinterpret_cast<struct pptp_msg*>(buffer);
    int ret = 0;

    switch (a_Client->m_CtrlState)
    {
	case CLIENT_WAITSTOPREPLY: // Client has requested a tear down of the connection
	    // This should never be reached
	    break;
	case CLIENT_ESTABLISHED: // Control connection has been established
	    switch (ntohs(pptp->ctrlType))
	    {
#if 0
		case SCCRQ: // Start-Control-Connection-Request [RFC2637 2.1]
		    if ((bufferUse == ntohs(pptp->length)) && (bufferUse == sizeof(struct sccrq_msg)) && (MAGIC_COOKIE == ntohl(pptp->magicCookie)))
		    {
			bufferUse = build_SCCRP(buffer, 1, 0, SCC_EXIST, 0, m_FramingCap, m_BearerCap,
				m_MaxChannels, m_FirmwareRev, static_cast<char*>(m_Hostname), static_cast<char*>(m_Vendor));
			a_Client->write(buffer, bufferUse);
		    }
		    else
		    {
			bufferUse = build_SCCRP(buffer, 1, 0, SCC_ERROR, BAD_FORMAT, 0, 0, 0, 0, NULL, NULL);
			a_Client->write(buffer, bufferUse);
		    }
		    break;
#endif
		case TCCRQ: // Stop-Control-Connection-Request [RFC2637 2.3]
		    if ((bufferUse == ntohs(pptp->length)) && (bufferUse == sizeof(struct tccrq_msg)) && (MAGIC_COOKIE == ntohl(pptp->magicCookie)))
		    {
			// Client requested to close the control connection
			// to which we immediately respond with OK. If this
			// session only has one client we also relay the
			// message to the server.
			// The client connection  should be closed after
			// sending the reply.
			if (1 == m_ClientList.size())
			{
			    m_Server->write(buffer, bufferUse);
			}
			bufferUse = build_TCCRP(buffer, TCC_OK, 0);
			a_Client->write(buffer, bufferUse);
			ret = -1;
		    }
		    break;
		case ECORQ: // Echo-Request [RFC2637 2.5]
		    if ((bufferUse == ntohs(pptp->length)) && (bufferUse == sizeof(struct ecorq_msg)) && (MAGIC_COOKIE == ntohl(pptp->magicCookie)))
		    {
			struct ecorp_msg* ecorp = reinterpret_cast<struct ecorp_msg*>(buffer);
			bufferUse = build_ECORP(buffer, ecorp->identifier, ECO_OK, 0);
			a_Client->write(buffer, bufferUse);
		    }
		    break;
		case ECORP: // Echo-Reply [RFC2637 2.6] */
		    if ((bufferUse == ntohs(pptp->length)) && (bufferUse == sizeof(struct ecorp_msg)) && (MAGIC_COOKIE == ntohl(pptp->magicCookie)))
		    {
			// TODO, Not implemented
		    }
		    break;
		case OCRQ: // Outgoing-Call-Request [RFC2637 2.7]
			if ((bufferUse == ntohs(pptp->length)) && (bufferUse == sizeof(struct ocrq_msg)) && (MAGIC_COOKIE == ntohl(pptp->magicCookie)))
			{
				struct ocrq_msg* ocrq = reinterpret_cast<struct ocrq_msg*>(buffer);
				Call* newCall;

				newCall = new Call(m_Server, a_Client);

				// Replace the ClientID with the ProxyID and relay to the server.
				newCall->m_ClientID = ocrq->callId;
				m_CallList.push_back(newCall);
				ocrq->callId = newCall->m_ProxyID;
				m_Server->write(buffer, bufferUse);
			}
			else
			{
				struct ocrq_msg* ocrq = reinterpret_cast<struct ocrq_msg*>(buffer);

				// The call request sent by the client was malformed so we reply
				// back to the client with a "bad format" error.
				bufferUse = build_OCRP(buffer, 0, ocrq->callId, OC_ERROR, BAD_FORMAT, 0, 0, 0, 0, 0);
				a_Client->write(buffer, bufferUse);
			}
			break;
		case CCRQ: // Call-Clear-Request [RFC2637 2.7]
			if ((bufferUse == ntohs(pptp->length)) && (bufferUse == sizeof(struct ccrq_msg)) && (MAGIC_COOKIE == ntohl(pptp->magicCookie)))
			{
			    struct ccrq_msg* ccrq = reinterpret_cast<struct ccrq_msg*>(buffer);

			    for (list<Call*>::iterator call_iter = m_CallList.begin(); call_iter != m_CallList.end(); call_iter++)
			    {
				if ((*call_iter)->m_ClientID == ccrq->callId)
				{
				    // Replace the client id with our proxy id and relay to the server.
				    ccrq->callId = (*call_iter)->m_ProxyID;
				    m_Server->write(buffer, bufferUse);

				    // Respond to the client with a Call Disconnect Notify message
				    bufferUse = build_CDN(buffer, (*call_iter)->m_ClientID, CDN_REQUEST, NONE, 0, NULL);
				    a_Client->write(buffer, bufferUse);

				    // Delete the call and also remove it from the call list
				    delete (*call_iter);
				    m_CallList.erase(call_iter);
				    break;
				}
			    }
			}
			break;
		case CDN: // Call-Disconnect-Notify [RFC2637 2.7]
			if ((bufferUse == ntohs(pptp->length)) && (bufferUse == sizeof(struct cdn_msg)) && (MAGIC_COOKIE == ntohl(pptp->magicCookie)))
			{
			    // TODO: implement CDN messages
#if 0
			    struct cdn_msg *cdn = reinterpret_cast<struct cdn_msg*>(mBuffer);

			    if (aClient->clientID() == cdn->callId)
			    {
				// We dont proxy call disconnect notifications yet.  Instead we just replace
				// the client id with our proxy id and relay to the server.
				cdn->callId = aClient->proxyID();
				mServer->write(mBuffer, mBufferUz);
			    }
#endif
			}
			break;
	    }
	    break;
	case CLIENT_WAITCTRLREPLY: // Control connection requested by the client
	    break;
	case CLIENT_IDLE: // No control connection established
	    switch (ntohs(pptp->ctrlType))
	    {
		case SCCRQ: // Start-Control-Connection-Request [RFC2637 2.1]
		    {
			struct sccrq_msg* sccrq = reinterpret_cast<struct sccrq_msg*>(buffer);

			if (bufferUse == ntohs(sccrq->length) && (bufferUse == sizeof(struct sccrq_msg)) && (MAGIC_COOKIE == ntohl(sccrq->magicCookie)))
			{
			    // TODO, verify that we still are accepting new connections
			    if (1)
			    {
				if (1 == m_ClientList.size())
				{ // Only the first client in this session may send SCCRQ to the server
				    Logger::getLogger()->log(LOG_INFO, "Accepted connection from %s:%u (%.3u/%.3u)",
					    inet_ntoa(a_Client->m_SockAddr4.sin_addr), a_Client->m_SockAddr4.sin_port, 1, 1);
				    bufferUse = build_SCCRQ(buffer, 1, 0, 3, 3, 0, 0, NULL, NULL);
				    m_Server->write(buffer, bufferUse);
				    a_Client->m_CtrlState = CLIENT_WAITCTRLREPLY;
				}
				else
				{ // All other clients will get a fake SCCRP message
				    bufferUse = build_SCCRP(buffer, 1, 0, SCC_OK, 0, m_FramingCap, m_BearerCap, m_MaxChannels, m_FirmwareRev, m_Hostname, m_Vendor);
				    a_Client->write(buffer, bufferUse);
				    a_Client->m_CtrlState = CLIENT_ESTABLISHED;
				}
			    }
			    else
			    {
				Logger::getLogger()->log(LOG_INFO, "Denied connection from %s:%u (No resource)",
					inet_ntoa(a_Client->m_SockAddr4.sin_addr), a_Client->m_SockAddr4.sin_port);
				bufferUse = build_SCCRP(buffer, 1, 0, SCC_ERROR, NO_RSRC, 0, 0, 0, 0, NULL, NULL);
				a_Client->write(buffer, bufferUse);
			    }
			}
			else
			{
			    bufferUse = build_SCCRP(buffer, 1, 0, SCC_ERROR, BAD_FORMAT, 0, 0, 0, 0, NULL, NULL);
			    a_Client->write(buffer, bufferUse);
			}
			break;
		    }
		default:
		    break;
	    }
	    break;
	default:
	    break;
    }

    return ret;
}

bool Session::operator==(const Server* a_Server)
{
    return !memcmp(&(m_Server->m_SockAddr4), &(a_Server->m_SockAddr4), sizeof(struct sockaddr_in));
}
