/* ----------------------------------------------------------------------------
   libconfig - A structured configuration file parsing library
   Copyright (C) 2005  Mark A Lindner
 
   This file is part of libconfig.
    
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.
    
   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
    
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
   ----------------------------------------------------------------------------
*/

%defines
%output="y.tab.c"
%pure-parser
%lex-param{void *scanner}
%parse-param{void *scanner}
%parse-param{struct parse_context *ctx}

%{
#include <stdio.h>
#include <string.h>
#include "libconfig.h"
#include "private.h"

static const char *err_array_elem_type = "mismatched element type in array";
static const char *err_duplicate_setting = "duplicate setting name";

#define IN_ARRAY()                              \
  (ctx->setting->type == CONFIG_TYPE_ARRAY)

void libconfig_yyerror(void *scanner, struct parse_context *ctx,
                      char const *s)
  {
  ctx->config->error_line = libconfig_yyget_lineno(scanner);
  ctx->config->error_text = s;
  }


%}

%union
  {
  long ival;
  double fval;
  char *sval;
  }

%token <ival> TOK_BOOLEAN TOK_INTEGER
%token <fval> TOK_FLOAT
%token <sval> TOK_STRING TOK_NAME
%token TOK_EQUALS TOK_NEWLINE TOK_ARRAY_START TOK_ARRAY_END TOK_COMMA TOK_GROUP_START TOK_GROUP_END TOK_END TOK_GARBAGE

%%

settings:
    /* empty */
  | settings setting
  ;
  
setting:
    assignment
  | group
  ;

array:
  TOK_ARRAY_START { ctx->setting->type = CONFIG_TYPE_ARRAY; }
  list
  TOK_ARRAY_END { }
  ;

value:
    simple_value
  | array
  ;

simple_value:
    TOK_BOOLEAN
  {
  if(IN_ARRAY())
    {
    if(! config_setting_set_bool_elem(ctx->setting, -1, (int)$1))
      {
      libconfig_yyerror(scanner, ctx, err_array_elem_type);
      YYABORT;
      }
    }
  else
    config_setting_set_bool(ctx->setting, (int)$1);
  }
  | TOK_INTEGER
  {
  if(IN_ARRAY())
    {
    if(! config_setting_set_int_elem(ctx->setting, -1, $1))
      {
      libconfig_yyerror(scanner, ctx, err_array_elem_type);
      YYABORT;
      }
    }
  else
    {
    config_setting_set_int(ctx->setting, $1);
    }
  }
  | TOK_FLOAT
  {
  if(IN_ARRAY())
    {
    if(! config_setting_set_float_elem(ctx->setting, -1, $1))
      {
      libconfig_yyerror(scanner, ctx, err_array_elem_type);
      YYABORT;
      }
    }
  else
    config_setting_set_float(ctx->setting, $1);
  }
  | TOK_STRING
  {
  if(IN_ARRAY())
    {
    if(! config_setting_set_string_elem(ctx->setting, -1, $1))
      {
      free($1);
      libconfig_yyerror(scanner, ctx, err_array_elem_type);
      YYABORT;
      }
    else
      free($1);
    }
  else
    {
    config_setting_set_string(ctx->setting, $1);
    free($1);
    }
  }
  ;

list:
    /* empty */
  | simple_value
  | list TOK_COMMA simple_value
  ;

assignment:
  TOK_NAME
  {
  ctx->setting = config_setting_add(ctx->group, $1, CONFIG_TYPE_NONE);
  free($1);
  
  if(ctx->setting == NULL)
    {
      libconfig_yyerror(scanner, ctx, err_duplicate_setting);
      YYABORT;
    }
  }

  TOK_EQUALS value TOK_END
  { ctx->setting = NULL; }
  ;

group:
  TOK_NAME
  {
  ctx->group = config_setting_add(ctx->group, $1, CONFIG_TYPE_GROUP);
  free($1);
  }
  TOK_GROUP_START settings TOK_GROUP_END
  {
  if(ctx->group)
    ctx->group = ctx->group->parent;
  }
  ;

%%
